var fs = require('fs'),
  util = require('util');

var async = require('async');
var tracey = require('tracey');

exports.Formatter = ProgressFormatter;

ProgressFormatter.init = function(emitter) {
  return new ProgressFormatter(emitter);
}
function ProgressFormatter(emitter) {
  emitter.on('suiteStart', this.suiteStart.bind(this));
  emitter.on('examplePass', this.examplePass.bind(this));
  emitter.on('examplePend', this.examplePend.bind(this));
  emitter.on('exampleFail', this.exampleFail.bind(this));
  emitter.on('exampleError', this.exampleError.bind(this));
  emitter.on('suiteComplete', this.suiteComplete.bind(this));

  this.output_stream = process.stdout;

  this.failures = [];
  this.pending = [];
  this.errors = [];
}

ProgressFormatter.prototype.write = function(str, wrap) {
  if (wrap)
    str = wrap(str);
  this.output_stream.write(str);
};
ProgressFormatter.prototype.writeln = function(str, wrap) {
  if (!str) str = '';
  this.write(str, wrap);
  this.write('\n');
};

ProgressFormatter.prototype.suiteStart = function() {
  this.startTime = new Date();
};

ProgressFormatter.prototype.examplePass = function(example) {
  this.write('.', green);
};
ProgressFormatter.prototype.examplePend = function(example, error) {
  this.write('*', yellow);
  this.pending.push({example: example, error: error});
};
ProgressFormatter.prototype.exampleFail = function(example, error) {
  this.write('F', red);
  this.failures.push({example: example, error: error});
};
ProgressFormatter.prototype.exampleError = function(example, error) {
  this.write('E', cyan);
  this.errors.push({example: example, error: error});
};
ProgressFormatter.prototype.suiteComplete = function(results) {
  this.timeTaken = new Date() - this.startTime;
  this.writeln('');
  async.series([
    this.writePending.bind(this),
    this.writeFailures.bind(this),
    this.writeErrors.bind(this),
    this.writeSummary.bind(this, results)
  ]);
};
ProgressFormatter.prototype.writePending = function(callback) {
  if (this.pending.length == 0) {
    callback();
    return;
  }
  this.writeln();
  this.writeln('Pending:', yellow);
  var i = 1;
  async.forEachSeries(this.pending,
    function iterator(pending, next) {
      var example = pending.example, error = pending.error;
      this.writeln();
      this.writeln('  ' + i + ') ' + example.full_description, yellow);
      if (error) {
        this.locateError(example, error, function(err, location, line) {
          this.writeln('     // ' + location, yellow);
          this.writeln('       ' + error.message, yellow);
          i += 1;
          next();
        }.bind(this));
      } else {
        var location = relativize_path(example.file_path) +
                 ':' + example.line_number;
        this.writeln('     // ' + location, yellow);
        this.writeln('       <unimplemented>', yellow);
        i += 1;
        next();
      }
    }.bind(this),
    callback
  );
};
ProgressFormatter.prototype.writeFailures = function(callback) {
  if (this.failures.length == 0) {
    callback();
    return;
  }
  this.writeln();
  this.writeln('Failures:', red);
  var i = 1;
  async.forEachSeries(this.failures,
    function iterator(failure, next) {
      var example = failure.example, error = failure.error;
      this.locateError(example, error, function(err, location, line) {
        this.writeln();
        this.writeln('  ' + i + ') ' + example.full_description, red);
        this.writeln('     // ' + location, red);
        if (line)
          this.writeln('     ' + line, red);
        if (error.message) {
          msg = error.message.replace(/\n/g, '\n       ');
          this.writeln('       ' + msg, red);
        } else {
          this.writeln('       expected: ' +
                 format_variable(error.expected), red);
          this.writeln('            got: ' +
                 format_variable(error.actual), red);
        }
        i += 1;
        next();
      }.bind(this));
    }.bind(this),
    callback
  );
};
ProgressFormatter.prototype.writeErrors = function(callback) {
  if (this.errors.length == 0) {
    callback();
    return;
  }
  this.writeln();
  this.writeln('Errors:', cyan);
  var i = 1;
  async.forEachSeries(this.errors,
    function iterator(item, next) {
      var example = item.example, error = item.error;
      this.locateError(example, error, function(err, location, line) {
        this.writeln();
        this.writeln('  ' + i + ') ' + example.full_description, cyan);
        this.writeln('     // ' + location, cyan);
        if (line)
          this.writeln('     ' + line, cyan);
        msg = error.toString().replace(/\n/g, '\n       ');
        this.writeln('     ' + msg, cyan);
        var stack_lines = relevant_trace_lines(example, error);
        for (var l in stack_lines) {
          this.writeln('       ' + stack_lines[l], cyan);
        }
        i += 1;
        next();
      }.bind(this));
    }.bind(this),
    callback
  );
};
ProgressFormatter.prototype.writeSummary = function(results) {
  this.writeln();
  this.write(results.total + ' spec' + (results.total != 1 ? 's' : ''));
  var bits = [];
  if (results.pass)
    bits.push(green(results.pass + ' passed'));
  if (results.pend)
    bits.push(yellow(results.pend + ' pending'));
  if (results.fail)
    bits.push(red(results.fail + ' failed'));
  if (results.error)
    bits.push(cyan(results.error + ' errored'));
  if (bits.length > 0) {
    this.write(' (' + bits.join(', ') + ')');
  }
  this.writeln();
  this.writeTimeTaken();
};

ProgressFormatter.prototype.writeTimeTaken = function() {
  var secs = this.timeTaken / 1000;
  this.writeln('Time Taken: ' + secs + 's');
};

ProgressFormatter.prototype.locateError = function(example, error, callback) {
  var line_match = find_failed_line(error, example.file_path);
  var location;
  if (!line_match) {
    location = relativize_path(example.file_path);
    location += ':' + example.line_number;
  } else {
    location = relativize_path(line_match.path);
    location += ':' + line_match.line;
  }
  if (callback.length == 2 || !line_match) {
    callback(null, location);
    return;
  }
  get_line(line_match.path, line_match.line, function(err, line) {
    if (err)
      callback(err);
    else
      callback(null, location, line);
  });
};

function relevant_trace_lines(example, error) {
  try {
    var stack = tracey(error);
  } catch (ex) {
    return [];
  }
  var relevant = [];
  stack.some(function(frame) {
    if (/nodespec\/lib/.test(frame.path)) {
      return true;
    }
    if (frame.path) {
      var line = frame.raw.replace(frame.path,
                     relativize_path(frame.path));
    } else {
      var line = frame.raw;
    }
    relevant.push(line);
    return false;
  });
  return relevant;
}

function find_failed_line(error, file) {
  if (!error.stack) {
    return null;
  }
  var stack = tracey(error);
  // Find the first line from the stack that is from file
  var match = null;
  stack.some(function(frame) {
    if (frame.path && ~frame.path.indexOf(file)) {
      match = frame;
    }
    return match;
  });
  return match;
}

function format_variable(variable) {
  return util.inspect(variable, false, null);
}

function relativize_path(path) {
  var cwd = process.cwd();
  return path.replace(cwd, '.');
}

function get_line(filename, line_no, callback) {
  line_no = parseInt(line_no);
  var data = fs.readFileSync(filename, 'utf8');
  var lines = data.split('\n');
  if (lines.length >= line_no) {
    callback(null, lines[line_no - 1].trim());
    return;
  }
  throw new Error('File end reached without finding line');
}

function colour(num, str) {
  return '\x1B[' + num + 'm' + (str ? str + '\x1B[39m' : '');
}

function green(str) {
  return colour(32, str);
}
function red(str) {
  return colour(31, str);
}
function yellow(str) {
  return colour(33, str);
}
function cyan(str) {
  return colour(36, str);
}
function normal(str) {
  return colour(39, str);
}
